@@ -1,29 +1,25 @@
-# import at_scat
-# import ftran
+from typing import Optional
 
 import astropy.io.fits as fits
 import astropy.units as u
 import gbmgeometry
 import numba as nb
 import numpy as np
-from numba import prange
-
-# TODO Remove this 
+from numba import prange, types
+# TODO Remove this
 from numba.typed import Dict
-from numba import types
+from responsum import InstrumentResponse
 
-try:
-    from threeML.utils.OGIP.response import InstrumentResponse
-
-except:
-    from responsum import InstrumentResponse
-
-from gbm_drm_gen.basersp_numba import get_database, get_trigdat_precalc_database
+from gbm_drm_gen.basersp_numba import (get_database,
+                                       get_trigdat_precalc_database)
+from gbm_drm_gen.input_edges import (BgoTTEEdges, InputEdges, NaiTTEEdges,
+                                     trigdat_edges, trigdat_out_edge,
+                                     tte_edges)
 from gbm_drm_gen.matrix_functions import (atscat_highres_ephoton_interpolator,
-                                          calc_sphere_dist, echan_integrator,
-                                          geo_to_space, geocoords,
-                                          highres_ephoton_interpolator, trfind,
-                                          closest)
+                                          calc_sphere_dist, closest,
+                                          echan_integrator, geo_to_space,
+                                          geocoords,
+                                          highres_ephoton_interpolator, trfind)
 from gbm_drm_gen.utils.geometry import ang2cart, is_occulted
 
 det_name_lookup = {
@@ -43,6 +39,23 @@ det_name_lookup = {
     "BGO_01": 13,
 }
 
+det_name_lookup2 = {
+    "n0": "NAI_00",
+    "n1": "NAI_01",
+    "n2": "NAI_02",
+    "n3": "NAI_03",
+    "n4": "NAI_04",
+    "n5": "NAI_05",
+    "n6": "NAI_06",
+    "n7": "NAI_07",
+    "n8": "NAI_08",
+    "n9": "NAI_09",
+    "na": "NAI_10",
+    "nb": "NAI_11",
+    "b0": "BGO_00",
+    "b1": "BGO_01"}
+
+
 
 lu = [
     "n0",
@@ -65,13 +78,13 @@ lu = [
 class DRMGen(object):
     def __init__(
         self,
-        quaternions,
-        sc_pos,
-        det_number,
+        position_interpolator: gbmgeometry.PositionInterpolator,
+        det_number: str,
         ebin_edge_in,
-        mat_type=0,
+        mat_type: int=0,
         ebin_edge_out=None,
-        occult=True,
+        occult:bool=True,
+        time:float=0,
     ):
         """
         A generic GBM DRM generator. This can be inherited for specific purposes.
@@ -111,30 +124,40 @@ class DRMGen(object):
         self._database_nb = get_database(lu[det_number])
         #######################################################################################
 
+        # this checks if the output edges do not change
+        
         self._trigdat = True
         self._trigdat_mask = np.array([], dtype=int)
+
         if det_number < 12:
-            trigdat_edges = np.array([3.4,10.,22.,44.,95.,300.,500.,800.,2000.])
+            trigdat_edges = np.array(
+                [3.4, 10., 22., 44., 95., 300., 500., 800., 2000.])
         else:
-            trigdat_edges = np.array([150.,400.,850.,1500.,3000.,5500.,10000.,20000.,50000.])
+            trigdat_edges = np.array(
+                [150., 400., 850., 1500., 3000., 5500., 10000., 20000., 50000.])
+
         for i, (e_l, e_h) in enumerate(zip(ebin_edge_out[:-1], ebin_edge_out[1:])):
             elow_index = np.argwhere(np.isclose(e_l,
                                                 trigdat_edges,
                                                 rtol=0.01))
-            if len(elow_index)==0:
-                    self._trigdat = False
-            elif np.isclose(e_h, trigdat_edges[elow_index[0,0]+1], rtol=0.01):
+            if len(elow_index) == 0:
+                self._trigdat = False
+            elif np.isclose(e_h, trigdat_edges[elow_index[0, 0]+1], rtol=0.01):
                 self._trigdat_mask = np.append(self._trigdat_mask,
-                                                   elow_index[0,0])
+                                               elow_index[0, 0])
             else:
                 self._trigdat = False
 
         if self._trigdat:
-            self._database_precalc_trigdat = get_trigdat_precalc_database(lu[det_number], self._trigdat_mask)
+            self._database_precalc_trigdat = get_trigdat_precalc_database(
+                lu[det_number], self._trigdat_mask)
+
         else:
             # dummy
-            self._database_precalc_trigdat = get_trigdat_precalc_database(lu[det_number], [0])
+            self._database_precalc_trigdat = get_trigdat_precalc_database(lu[det_number], [
+                                                                          0])
             self._trigdat_mask = None
+
         #######################################################################################
         self._ein = np.zeros(self._nobins_in, dtype=np.float32)
         energ_lo = self._database_nb.energ_lo
@@ -147,14 +170,248 @@ class DRMGen(object):
 
         self._nobins_out = len(self._out_edge) - 1
 
-        self._quaternions = quaternions
-        self._sc_pos = sc_pos
 
-        self._compute_spacecraft_coordinates()
+        self._position_interpolator = position_interpolator
 
-    # @classmethod
-    # def from_tte(cls)
+        self.set_time(time)
+        
 
+    @classmethod
+    def from_128_bin_data(
+            cls,
+            det_name,
+            time: float = 0.0,
+            cspecfile: Optional[str] = None,
+            trigdat: Optional[str] = None,
+            poshist: Optional[str] = None,
+            T0: Optional[float] = None,
+            mat_type: int = 0,
+            custom_input_edges: Optional[InputEdges] = None,
+            occult: bool = False,
+    ):
+        """
+        A TTE/CSPEC specific drmgen already incorporating the standard input edges. Output edges are obtained
+        from lib import funs the input cspec file. Spacecraft position is read from the TTE file. For further details see the 
+        generic reader (DRMGen).
+
+        :param det_name: either NAI_{**} or BGO_{**} or n* b*
+        :param trigdat: the path to a trigdat file
+        :param mat_type: 0=direct 1=scattered 2=direct+scattered
+        :param time: time relative to trigger to pull spacecraft position or MET if using a poshist file
+        :param cspecfile: the cspecfile to pull energy output edges from
+        :param poshist: read a poshist file
+        """
+
+
+        if det_name not in det_name_lookup:
+
+            if det_name not in det_name_lookup2:
+
+                raise RuntimeError(f"{det_name} is not valid")
+
+            else:
+
+                det_name = det_name_lookup2[det_name]
+
+        det_number = det_name_lookup[det_name]
+
+        if det_number > 11:
+            # BGO
+
+            if custom_input_edges is None:
+
+                in_edge = tte_edges["bgo"]
+
+            else:
+
+                assert isinstance(
+                    custom_input_edges, BgoTTEEdges), f"custom edges are not an instance of BgoTTEEdges!"
+
+                in_edge = custom_input_edges.edges
+
+        else:
+
+            if custom_input_edges is None:
+
+                in_edge = tte_edges["nai"]
+
+            else:
+
+                assert isinstance(
+                    custom_input_edges, NaiTTEEdges), f"custom edges are not an instance of NaiTTEEdges!"
+
+                in_edge = custom_input_edges.edges
+
+        # Create the out edge energies
+        with fits.open(cspecfile) as f:
+            out_edge = np.zeros(129, dtype=np.float32)
+            out_edge[:-1] = f["EBOUNDS"].data["E_MIN"]
+            out_edge[-1] = f["EBOUNDS"].data["E_MAX"][-1]
+
+        out_edge = out_edge
+
+        if trigdat is not None:
+
+            try:
+                position_interpolator = gbmgeometry.PositionInterpolator.from_trigdat(
+                    trigdat_file=trigdat
+                )
+
+            except:
+
+                position_interpolator = gbmgeometry.PositionInterpolator.from_trigdat_hdf5(
+                    trigdat_file=trigdat
+                )
+
+            gbm = gbmgeometry.GBM(
+                position_interpolator.quaternion(time),
+                position_interpolator.sc_pos(time) * u.km,
+            )
+
+        elif poshist is not None:
+
+            try:
+
+                position_interpolator = gbmgeometry.PositionInterpolator.from_poshist(
+                    poshist_file=poshist, T0=T0
+                )
+
+            except:
+
+                position_interpolator = gbmgeometry.PositionInterpolator.from_poshist_hdf5(
+                    poshist_file=poshist, T0=T0
+                )
+
+            gbm = gbmgeometry.GBM(
+                position_interpolator.quaternion(time),
+                position_interpolator.sc_pos(time) * u.m,
+            )
+
+        else:
+
+            raise RuntimeError("No trigdat or posthist file used!")
+
+        out = cls(
+            position_interpolator=position_interpolator,
+            det_number=det_number,
+            ebin_edge_in=in_edge,
+            mat_type=mat_type,
+            ebin_edge_out=out_edge,
+            occult=occult,
+            time=time
+        )
+
+        out._gbm = gbm
+
+        return out
+
+    @classmethod
+    def from_trigdat(cls,
+                     trigdat_file: str,
+                     det,
+                     mat_type=0,
+                     tstart=0,
+                     tstop=0.0,
+                     time=0.0,
+                     occult=False):
+
+        """TODO describe function
+
+        :param cls: 
+        :type cls: 
+        :param trigdat_file: 
+        :type trigdat_file: str
+        :param det: 
+        :type det: 
+        :param mat_type: 
+        :type mat_type: 
+        :param tstart: 
+        :type tstart: 
+        :param tstop: 
+        :type tstop: 
+        :param time: 
+        :type time: 
+        :param occult: 
+        :type occult: 
+        :returns: 
+
+        """
+
+        matrix_type = mat_type
+
+        
+        det_number = det
+
+        time = time
+
+        maxen = 140
+
+        
+        # Setup the input side energy edges
+        if det > 11:
+
+            in_edge = trigdat_edges["bgo"]
+            out_edge = trigdat_out_edge["bgo"]
+
+        else:
+
+            in_edge = trigdat_edges["nai"]
+            out_edge = trigdat_out_edge["nai"]
+
+
+        try:
+            position_interpolator = gbmgeometry.PositionInterpolator.from_trigdat(
+                trigdat_file=trigdat_file
+            )
+
+        except:
+
+            position_interpolator = gbmgeometry.PositionInterpolator.from_trigdat_hdf5(
+                trigdat_file=trigdat_file
+                )
+
+        out = cls(
+            position_interpolator=position_interpolator,
+            det_number=det_number,
+            ebin_edge_in=in_edge,
+            mat_type=matrix_type,
+            ebin_edge_out=out_edge,
+            occult=occult,
+            time = time
+        )
+
+        out._tstart = tstart
+        out._tstop = tstop
+
+        return out
+
+
+
+    @property
+    def postion_interpolator(self) -> gbmgeometry.PositionInterpolator:
+
+        return self._position_interpolator
+
+    @property
+    def current_met(self) -> float:
+
+        return self._position_interpolator.met(self._time)
+
+    def met_at(self, t: float) -> float:
+
+        """
+        returns the MET at a given relative time
+
+        :param t: 
+        :type t: float
+        :returns: 
+
+        """
+        return self._position_interpolator.met(t)
+    
+
+        
+    
     @property
     def ebounds(self):
 
@@ -170,12 +427,29 @@ class DRMGen(object):
 
         return self._drm.T
 
-    def to_3ML_response(self, ra, dec):
+    def to_3ML_response(self, ra, dec, coverage_interval=None):
+        """
+        create an instrument reponse object for 3ML
+        
+        :param ra: 
+        :type ra: 
+        :param dec: 
+        :type dec: 
+        :returns: 
 
-        self.set_location(ra, dec, use_numba=True)
+        """
+        self.set_location(ra, dec)
+
+        out = self.matrix
+
+        if not np.all(np.isfinite(out)):
+
+            for i, j in zip(np.where(np.isnan(out))[0], np.where(np.isnan(out))[1]):
+
+                out[i, j] = 0.
 
         response = InstrumentResponse(
-            self.matrix, self.ebounds, self.monte_carlo_energies
+            out, self.ebounds, self.monte_carlo_energies, coverage_interval=coverage_interval
         )
 
         return response
@@ -205,7 +479,7 @@ class DRMGen(object):
                            ), "GLAST", "GBM", overwrite
         )
 
-    def set_location(self, ra, dec, use_numba=True):
+    def set_location(self, ra, dec) -> None:
         """
         Set the Ra and Dec of the DRM to be built. This invokes DRM generation as well.
 
@@ -219,20 +493,23 @@ class DRMGen(object):
         if self._occult:
             if is_occulted(ra, dec, self._sc_pos):
                 self._drm = self._occulted_DRM
-                return None
-                
-        # get the spacecraft coordinates
-        az, el = self._get_coords(ra, dec)
+            else:
+                az, el = self._get_coords(ra, dec)
+                self._drm = self._make_drm_numba(
+                    az, el, self._geo_az, self._geo_el)
+        else:
+            # get the spacecraft coordinates
+            az, el = self._get_coords(ra, dec)
 
-        self._drm = self._make_drm_numba(
-            az, el, self._geo_az, self._geo_el)
+            self._drm = self._make_drm_numba(
+                az, el, self._geo_az, self._geo_el)
 
         # go ahead and transpose it for spectal fitting, etc.
         # self._drm_transpose = self._drm.T
         # go ahead and transpose it for spectal fitting, etc.
         # self._drm_transpose = self._drm.T
 
-    def set_location_direct_sat_coord(self, az, el):
+    def set_location_direct_sat_coord(self, az, el) -> None:
         """
         Set the AZ and EL in satellite coordinates of the DRM to be built. This invokes DRM generation as well.
 
@@ -262,9 +539,14 @@ class DRMGen(object):
 
         self._sc_quaternions_updater()
 
+        self._compute_spacecraft_coordinates()
+        
+
     def _sc_quaternions_updater(self):
 
-        raise NotImplementedError("implemented in subclass")
+        self._quaternions = self._position_interpolator.quaternion(self._time)
+
+        self._sc_pos = self._position_interpolator.sc_pos(self._time)
 
     def _compute_spacecraft_coordinates(self):
         """
@@ -465,7 +747,7 @@ def _build_drm(
     i1 += 1
     i2 += 1
     i3 += 1
-    
+
     # mod here
 
     mat1 = rsps[milliaz[i1 - 1] + "_" + millizen[i1 - 1]]
@@ -518,11 +800,11 @@ def _build_drm(
             if lat <= lat_cent[0]:
                 il_low = 0
                 il_high = 1
-                #TODO Changes this. Was 0.
+                # TODO Changes this. Was 0.
                 l_frac = 1.0
             else:
                 idx = np.where(lat_cent < lat)[0][-1]
-                #TODO Changed this. This was il_low = idx -1 and il_high = idx
+                # TODO Changed this. This was il_low = idx -1 and il_high = idx
                 il_low = idx
                 il_high = idx+1
                 l_frac = 1.0 - (lat - lat_cent[idx]) / (
@@ -541,7 +823,7 @@ def _build_drm(
             sf = np.arctan(1.0) / 45.0
 
             at_scat(tmp_out, num_theta, num_phi, theta_u, phi_u, gx, gy, gz,
-            sf, grid_points_list, trigdat_precalc_rsps, rsps, milliaz, millizen, epx_lo, epx_hi, ichan, out_edge,
+                    sf, grid_points_list, trigdat_precalc_rsps, rsps, milliaz, millizen, epx_lo, epx_hi, ichan, out_edge,
                     at_scat_data, il_low, il_high, l_frac, trigdat)
             tmp_out2 = np.zeros((ienerg, nobins_out))
             for i in range(num_theta):
@@ -687,13 +969,14 @@ def _build_drm(
     ) / 2.0
     return final_drm
 
+
 @nb.njit(parallel=True, fastmath=True)
 def at_scat(tmp_out, num_theta, num_phi, theta_u, phi_u, gx, gy, gz,
             sf, grid_points_list, trigdat_precalc_rsps, rsps, milliaz, millizen, epx_lo, epx_hi, ichan, out_edge,
             at_scat_data, il_low, il_high, l_frac, trigdat):
     for i in prange(num_theta):
         for j in prange(num_phi):
-            #TODO: Changed this. This was range(1), but I think 2 is needed to also cover
+            # TODO: Changed this. This was range(1), but I think 2 is needed to also cover
             # the negative phi values
             for k in prange(2):
                 dirx, diry, dirz, az, el = geo_to_space(
@@ -716,29 +999,31 @@ def at_scat(tmp_out, num_theta, num_phi, theta_u, phi_u, gx, gy, gz,
                 i1 = closest(P, grid_points_list)
 
                 if trigdat:
-                    rsps_echan_integrator = trigdat_precalc_rsps[milliaz[i1] + "_" + millizen[i1]]#i1
+                    # i1
+                    rsps_echan_integrator = trigdat_precalc_rsps[milliaz[i1] +
+                                                                 "_" + millizen[i1]]
 
                 else:
-                    rsps_echan_integrator = echan_integrator(rsps[milliaz[i1] + "_" + millizen[i1]],#i1
+                    rsps_echan_integrator = echan_integrator(rsps[milliaz[i1] + "_" + millizen[i1]],  # i1
                                                              epx_lo,
                                                              epx_hi,
                                                              ichan,
                                                              out_edge
-                    )
+                                                             )
 
                 msum(at_scat_data[..., il_low, i, j],
                      at_scat_data[..., il_high, i, j],
                      np.ascontiguousarray(rsps_echan_integrator),
-                     tmp_out[i,j,k],
+                     tmp_out[i, j, k],
                      l_frac,
-                )
+                     )
+
 
-                
 @nb.njit(fastmath=True, parallel=False)
 def msum(this_data_lo, this_data_hi, direct_diff_matrix, tmp_out, l_frac):
     tmp = this_data_lo*l_frac+this_data_hi*(1-l_frac)
     tmp_out += np.dot(tmp, direct_diff_matrix)
-    #for ii in prange(ienerg):
+    # for ii in prange(ienerg):
     #    for jj in prange(nobins_out):
     #        for kk in prange(ienerg):
     #            tmp_out[ii, jj] += (
